import { None } from './option.js'

class UnwrapError extends Error {
  constructor(message) {
    super(message)
    this.name = 'UnwrapError'
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, UnwrapError)
    }
  }
}

export class Result {
  static ok(value) {
    return new Ok(value)
  }

  static err(error) {
    return new Err(error)
  }

  isOk() {
    return this instanceof Ok
  }

  isErr() {
    return this instanceof Err
  }
}

export class Ok extends Result {
  #value

  constructor(value) {
    super()
    this.#value = value
    Object.freeze(this)
  }

  ap(value) {
    return this.#value(value)
  }

  andThen(fn) {
    return fn(this.#value)
  }

  map(fn) {
    return new Ok(fn(this.#value))
  }

  filterOr(predicate, err) {
    if (predicate(this.#value)) {
      return this
    } else {
      return Result.err(err)
    }
  }

  filterOrElse(predicate, fn) {
    if (predicate(this.#value)) {
      return this
    } else {
      return Result.err(fn(this.#value))
    }
  }

  inspect(fn) {
    fn(this.#value)
    return this
  }

  ok() {
    return Option.some(this.#value)
  }

  err() {
    return None
  }

  orOther(_value) {
    return this
  }

  orElse(_fn) {
    return this
  }

  unwrap() {
    return this.#value
  }

  unwrapErr() {
    throw new UnwrapError('attempted to unwrapErr an Ok value')
  }

  toString() {
    return `Ok(${this.#value})`
  }
}

export class Err extends Result {
  #error

  constructor(error) {
    super()
    this.#error = error
    Object.freeze(this)
  }

  ap(_value) {
    return this
  }

  andThen(_fn) {
    return this
  }

  map(_fn) {
    return this
  }

  filterOr(_predicate, _err) {
    return this
  }

  filterOrElse(_predicate, _fn) {
    return this
  }

  inspect(_fn) {
    return this
  }

  ok() {
    return None
  }

  err() {
    return Option.some(this.#error)
  }

  orOther(value) {
    return value
  }

  orElse(fn) {
    return fn(this.#error)
  }

  unwrap() {
    throw new UnwrapError('attempted to unwrap an Err value')
  }

  unwrapErr() {
    return this.#error
  }

  toString() {
    return `Err(${this.#error})`
  }
}

